part of apptive_grid_model;

/// The Data Types that are available in ApptiveGrid
enum DataType {
  /// Type to display [String]
  text(backendName: 'string'),

  /// Type to display [DateTime]
  dateTime(backendName: 'date-time'),

  /// Type to display [DateTime] without the option to adjust the Time part
  date(backendName: 'date'),

  /// Type to display [int] numbers
  integer(backendName: 'integer'),

  /// Type to display [double] decimal numbers
  decimal(backendName: 'decimal'),

  /// Type to display [bool] values
  checkbox(backendName: 'boolean'),

  /// Type to display enum values
  singleSelect(backendName: 'enum'),

  /// Select Multiple Values from List of possible Values
  enumCollection(backendName: 'enumcollection'),

  /// Type to display CrossReference Values
  crossReference(backendName: 'reference'),

  /// Type for Attachments
  attachment(backendName: 'attachments'),

  /// Type for geolocations
  geolocation(backendName: 'geolocation'),

  /// Type for Multi Cross References
  multiCrossReference(backendName: 'references'),

  /// Type for UserReferences. Used for createdBy
  userReference(backendName: 'createdby');

  /// Define a Datatype with a corresponding [backendName]
  const DataType({
    required this.backendName,
  });

  /// The name that is used in the ApptiveGridBackend for this [DataType]
  final String backendName;
}

/// Returns [DataType] that matching a certain schema [schemaProperty]
///
/// throws [ArgumentError] if DataType is not supported yet
DataType dataTypeFromSchemaProperty({
  required dynamic schemaProperty,
}) {
  final schemaType = schemaProperty['type'];
  final format = schemaProperty['format'];
  switch (schemaType) {
    case 'string':
      if (schemaProperty['enum'] != null) {
        return DataType.singleSelect;
      }
      switch (format) {
        case 'date-time':
          return DataType.dateTime;
        case 'date':
          return DataType.date;
      }
      return DataType.text;
    case 'integer':
      return DataType.integer;
    case 'number':
      return DataType.decimal;
    case 'boolean':
      return DataType.checkbox;
    case 'object':
      final objectType = schemaProperty['objectType'];
      switch (objectType) {
        case 'entityreference':
          return DataType.crossReference;
        case 'attachment':
          return DataType.attachment;
        case 'geolocation':
          return DataType.geolocation;
        case 'userReference':
          return DataType.userReference;
        default:
          throw ArgumentError(
            'No defined Object type for type: $objectType',
          );
      }
    case 'array':
      final itemType =
          dataTypeFromSchemaProperty(schemaProperty: schemaProperty['items']);
      switch (itemType) {
        case DataType.attachment:
          return DataType.attachment;
        case DataType.singleSelect:
          return DataType.enumCollection;
        case DataType.crossReference:
          return DataType.multiCrossReference;
        default:
          throw ArgumentError(
            'No defined Array type for type: $itemType',
          );
      }
  }
  throw ArgumentError(
    'No according DataType found for "$schemaType". Supported DataTypes are ${DataType.values}',
  );
}
